"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Certificate = void 0;
const ethers_1 = require("ethers");
const events_1 = require("../utils/events");
const CertificateUtils_1 = require("./CertificateUtils");
const CertificationRequest_1 = require("./CertificationRequest");
const PreciseProofUtils_1 = require("../utils/PreciseProofUtils");
class Certificate {
    constructor(id, blockchainProperties) {
        this.id = id;
        this.blockchainProperties = blockchainProperties;
        this.initialized = false;
    }
    static async create(to, value, fromTime, toTime, deviceId, blockchainProperties) {
        if (value.gt(CertificationRequest_1.MAX_ENERGY_PER_CERTIFICATE)) {
            throw new Error(`Too much energy requested. Requested: ${value}, Max: ${CertificationRequest_1.MAX_ENERGY_PER_CERTIFICATE}`);
        }
        const newCertificate = new Certificate(null, blockchainProperties);
        const getIdFromEvents = (logs) => Number(logs.find((log) => log.event === 'CertificationRequestApproved').topics[2]);
        const { issuer } = blockchainProperties;
        const issuerWithSigner = issuer.connect(blockchainProperties.activeUser);
        const data = await issuer.encodeData(fromTime, toTime, deviceId);
        const properChecksumToAddress = ethers_1.ethers.utils.getAddress(to);
        const tx = await issuerWithSigner.issue(properChecksumToAddress, value, data);
        const { events } = await tx.wait();
        newCertificate.id = getIdFromEvents(events);
        return newCertificate.sync();
    }
    static async createPrivate(to, value, fromTime, toTime, deviceId, blockchainProperties) {
        if (value.gt(CertificationRequest_1.MAX_ENERGY_PER_CERTIFICATE)) {
            throw new Error(`Too much energy requested. Requested: ${value}, Max: ${CertificationRequest_1.MAX_ENERGY_PER_CERTIFICATE}`);
        }
        const newCertificate = new Certificate(null, blockchainProperties);
        const getIdFromEvents = (logs) => Number(logs.find((log) => log.event === 'CertificationRequestApproved').topics[2]);
        const { issuer } = blockchainProperties;
        const issuerWithSigner = issuer.connect(blockchainProperties.activeUser);
        const data = await issuer.encodeData(fromTime, toTime, deviceId);
        const properChecksumToAddress = ethers_1.ethers.utils.getAddress(to);
        const ownershipCommitment = {
            [properChecksumToAddress]: value.toString()
        };
        const commitmentProof = PreciseProofUtils_1.PreciseProofUtils.generateProofs(ownershipCommitment);
        const tx = await issuerWithSigner.issuePrivate(properChecksumToAddress, commitmentProof.rootHash, data);
        const { events } = await tx.wait();
        newCertificate.id = getIdFromEvents(events);
        return {
            certificate: await newCertificate.sync(),
            proof: {
                ...commitmentProof,
                txHash: tx.hash
            }
        };
    }
    async sync() {
        if (this.id === null) {
            return this;
        }
        const { registry } = this.blockchainProperties;
        const certOnChain = await registry.getCertificate(this.id);
        this.data = certOnChain.data;
        const { issuer } = this.blockchainProperties;
        const decodedData = await issuer.decodeData(this.data);
        const allIssuanceLogs = await events_1.getEventsFromContract(registry, registry.filters.IssuanceSingle(null, null, null));
        const issuanceLog = allIssuanceLogs.filter((event) => event._id.toString() === this.id.toString())[0];
        const issuanceBlock = await registry.provider.getBlock(issuanceLog.blockHash);
        this.generationStartTime = Number(decodedData['0']);
        this.generationEndTime = Number(decodedData['1']);
        this.deviceId = decodedData['2'];
        this.issuer = certOnChain.issuer;
        this.creationTime = Number(issuanceBlock.timestamp);
        this.creationBlockHash = issuanceLog.blockHash;
        const certificationRequestApprovedEvents = await events_1.getEventsFromContract(issuer, issuer.filters.CertificationRequestApproved(null, null, this.id));
        this.certificationRequestId = certificationRequestApprovedEvents[0]._id;
        this.owners = await CertificateUtils_1.calculateOwnership(this.id, this.blockchainProperties);
        this.claimers = await CertificateUtils_1.calculateClaims(this.id, this.blockchainProperties);
        this.initialized = true;
        return this;
    }
    async claim(claimData, amount, from, to) {
        var _a;
        const { activeUser, registry } = this.blockchainProperties;
        const registryWithSigner = registry.connect(activeUser);
        const activeUserAddress = await activeUser.getAddress();
        const claimAddress = to !== null && to !== void 0 ? to : activeUserAddress;
        const ownedVolume = ethers_1.BigNumber.from((_a = this.owners[claimAddress]) !== null && _a !== void 0 ? _a : 0);
        if (ownedVolume.eq(0) && !amount) {
            throw new Error(`claim(): ${claimAddress} does not own a share in the certificate.`);
        }
        const encodedClaimData = await CertificateUtils_1.encodeClaimData(claimData, this.blockchainProperties);
        const claimTx = await registryWithSigner.safeTransferAndClaimFrom(from !== null && from !== void 0 ? from : activeUserAddress, claimAddress, this.id, amount !== null && amount !== void 0 ? amount : ownedVolume, this.data, encodedClaimData);
        const txReceipt = await claimTx.wait();
        return txReceipt;
    }
    async transfer(to, amount, from) {
        var _a;
        if (await this.isRevoked()) {
            throw new Error(`Unable to transfer Certificate #${this.id}. It has been revoked.`);
        }
        const { activeUser, registry } = this.blockchainProperties;
        const fromAddress = from !== null && from !== void 0 ? from : (await activeUser.getAddress());
        const toAddress = ethers_1.ethers.utils.getAddress(to);
        const ownedVolume = ethers_1.BigNumber.from((_a = this.owners[toAddress]) !== null && _a !== void 0 ? _a : 0);
        const registryWithSigner = registry.connect(activeUser);
        const tx = await registryWithSigner.safeTransferFrom(fromAddress, toAddress, this.id, amount !== null && amount !== void 0 ? amount : ownedVolume, this.data);
        await tx.wait();
        return tx;
    }
    async revoke() {
        const { issuer } = this.blockchainProperties;
        const issuerWithSigner = issuer.connect(this.blockchainProperties.activeUser);
        const tx = await issuerWithSigner.revokeCertificate(this.id);
        await tx.wait();
        return tx;
    }
    async isRevoked() {
        const { issuer } = this.blockchainProperties;
        const revokedEvents = await events_1.getEventsFromContract(issuer, issuer.filters.CertificateRevoked(this.id));
        return revokedEvents.length > 0;
    }
    async getClaimedData() {
        var _a;
        const { registry } = this.blockchainProperties;
        const claims = [];
        const claimSingleEvents = await events_1.getEventsFromContract(registry, registry.filters.ClaimSingle(null, null, null, null, null, null));
        for (const claimEvent of claimSingleEvents) {
            if (claimEvent._id.toNumber() === this.id) {
                const { _claimData, _id, _claimIssuer, _claimSubject, _topic, _value } = claimEvent;
                const claimData = await CertificateUtils_1.decodeClaimData(_claimData, this.blockchainProperties);
                claims.push({
                    id: _id.toNumber(),
                    from: _claimIssuer,
                    to: _claimSubject,
                    topic: _topic.toNumber(),
                    value: _value.toNumber(),
                    claimData
                });
            }
        }
        const claimBatchEvents = await events_1.getEventsFromContract(registry, registry.filters.ClaimBatch(null, null, null, null, null, null));
        for (const claimBatchEvent of claimBatchEvents) {
            if (claimBatchEvent._ids.map((idAsBN) => idAsBN.toNumber()).includes(this.id)) {
                const { _ids, _claimData, _claimIssuer, _claimSubject, _topics, _values } = claimBatchEvent;
                const claimIds = _ids.map((idAsBN) => idAsBN.toNumber());
                const index = claimIds.indexOf(this.id);
                const claimData = await CertificateUtils_1.decodeClaimData(_claimData[index], this.blockchainProperties);
                claims.push({
                    id: _ids[index].toNumber(),
                    from: _claimIssuer,
                    to: _claimSubject,
                    topic: (_a = _topics[index]) === null || _a === void 0 ? void 0 : _a.toNumber(),
                    value: _values[index].toNumber(),
                    claimData
                });
            }
        }
        return claims;
    }
}
exports.Certificate = Certificate;
//# sourceMappingURL=Certificate.js.map